"""

This file is part of pyEnsembl.
    Copyright (C) 2018, Andrés García

    Any questions, comments or issues can be addressed to a.garcia230395@gmail.com.

"""

from .restclient import RESTClient, HTTPError
import pprint


class BaseEnsemblRESTClient:
    """Base client for an Ensembl REST API."""
    
    def __init__(self, base_url=None):
        if base_url is None:
            base_url = self.base_url
            
        self.rest_client = RESTClient(base_url)
    # ----
        
    def make_request(self, resource, *args, params=None, headers=None, **kwargs):
        "Follow the route mapping the arguments to the correct place."
        # Assemble the request
        request_type, route_template = resource.split(' ')
        route = self._map_arguments(route_template, *args, **kwargs)
        
        endpoint = self.rest_client.route(*route)
        
        # Perform the request
        while True:
            try:
                return endpoint.do(request_type, params, headers=headers)

            # Handle rate limit
            except HTTPError as e:
                    if e.response.status_code == 429:
                        # Maximum requests rate exceded
                        # Need to wait
                        response = e.response
                        wait_time = float(response.headers["Retry-After"])
                        sys.stderr.write('Maximum requests limit reached, waiting for ' 
                                         + str(wait_time)
                                         + 'secs')
                        time.sleep(wait_time)
                    else:
                        raise
    # ---        
        
    def _map_arguments(self, route, *args, **kwargs):
        """Map the arguments to the template.
        
        The template is a string of the form:
        "some/:string/with/:semicolons"
        
        First maps the **kwargs, then maps the *args in order.
        If this fails throw an exception.
        
        """
        # First split into arguments: 
        # "some/:string/with/:semicolons" -> ['some', ':string', 'with', ':semicolons'] 
        parts = route.split('/')
        # Then replace the semicolons with format strings
        parts = ['{' + s[1:] + '}' if s.startswith(':') else s for s in parts]
        
        # Then map the keyword arguments
        parts = [self._format_if_possible(s, **kwargs) for s in parts]
        
        # Finally, map the positional arguments
        args_iter = iter(args)
        mapped_parts = []
        for s in parts:
            if s.startswith('{'):
                try:
                    mapped_parts.append(next(args_iter))
                except StopIteration:
                    # Parameter not provided, skip in case it is optional
                    pass
            else:
                mapped_parts.append(s)
        
        return mapped_parts
    # ---
           
    def _format_if_possible(self, format_string, **kwargs):
        """Try to apply the arguments to the format string.
        
        If not possible, return the string unchanged.
        """
        try:
            return format_string.format(**kwargs)
        except KeyError:
            return format_string
    # ---
    
# --- BaseClient




## Now are the factories for the automated creation of the classes
##  Y Y Y Y Y
##  | | | | |
##  v v v v v

def _format_parameters_docstring(parameters):
    fields = 'Name,Type,Description,Default,Example Values'.split(',')

    return (
        '\n\n'.join(
            '\n'.join(f'\t+ *{"*"+field+"*" if "Name" in field else field}*:  {parameter[field]}' for field in fields)
                for parameter in parameters)
    )

def _endpoint_docstring(endpoint):
    resource_info = '\n'.join(f'- **{name}**:\t{values}' for name, values in endpoint['resource_info'].items())
    return (
        f"{endpoint['category']} ``{endpoint['resource_string']}``\n\n"
        f"{endpoint['description']}\n\n\n"
        f"**Parameters**\n\n"
        f"- Required:\n{_format_parameters_docstring(endpoint['parameters']['required'])}\n\n"
        f"- Optional:\n{_format_parameters_docstring(endpoint['parameters']['optional'])}\n\n\n"
        f"**Resource info**\n\n"
        f"{resource_info}\n\n\n"
        f"**More info**\n\n"
        f"{endpoint['documentation_url']}\n\n"
    )
# ---

def _create_method(method_name, endpoint):
    """Create a class method"""
    
    def method(self, *args, **kwargs):
        return self.make_request(endpoint['resource_string'], *args, **kwargs)
    # ---
    
    method.__doc__ = _endpoint_docstring(endpoint)
    method.__name__ = method_name
    
    return method
# ---

def build_client_class(name, api_table, base_url, doc=''):
    """Create a new class that implements the methods of the API."""
    # Create the class dictionary
    class_dict = {'__doc__': doc,
                  'base_url': base_url}
    
    # Create the class methods
    methods = {endpoint['name'] : _create_method(endpoint['name'], endpoint) 
               for endpoint in api_table}
    
    class_dict.update(methods)
    
    # Create the class (a subclass of the BaseEnsemblRESTClient)
    return type(name, (BaseEnsemblRESTClient,), class_dict)
# ---



"""

This file is part of pyEnsemblRest.
    Copyright (C) 2018, Andrés García

    Any questions, comments or issues can be addressed to a.garcia230395@gmail.com.

"""

import json
from importlib_resources import open_text, path
from .core import build_client_class

# Import API information
with open_text('ensembl_rest.data','ensembl_rest_endpoints.json') as f:
    _api_table = json.load(f)


# Now create the main classes 
ensembl_rest_url = 'https://rest.ensembl.org'

EnsemblClient = build_client_class(
                'EnsemblClient', 
                _api_table,
                base_url=ensembl_rest_url,
                doc=f'A client for the Ensembl REST API ({ensembl_rest_url})')


def region_str(chrom, start, end=None, strand=+1):
    """Assemble a region string suitable for consumption for the Ensembl REST API.
    
    The generated string has the format: ``{chrom}:{start}..{end}:{strand}``
    
    """
    if end is None:
        end = start
        
    return f'{chrom}:{start}..{end}:{strand}'
# --- 



"""

This file is part of pyEnsembl.
    Derived from a file in the RESTEasy project by Arijit Basu <sayanarijit@gmail.com>.
    Copyright (C) 2018, Andrés García

    Any questions, comments or issues can be addressed to a.garcia230395@gmail.com.

"""

from __future__ import absolute_import, unicode_literals
import json
import requests
from simplejson.decoder import JSONDecodeError
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


class HTTPError(Exception):
    def __init__(self, response):
        self.response = response
        try:
            Exception.__init__(self, 'Server returned HTTP status code: %s\nContent: %s' % (response.status_code, response.json()))
        except JSONDecodeError:
            Exception.__init__(self, 'Server returned HTTP status code: %s\nContent: %s' % (response.status_code, response.text))

class InvalidResponseError(Exception):
    def __init__(self, response):
        self.response = response
        try:
            Exception.__init__(self, 'Server returned incompatible response:\n', response.json())
        except JSONDecodeError:
            Exception.__init__(self, 'Server returned incompatible response:\n', response.text)


class RESTClient:
    """
    REST session creator
    """
    def __init__(self, base_url, auth=None, verify=False, cert=None, timeout=None,
                 encoder=json.dumps, decoder=json.loads, debug=False):
        self.base_url = base_url
        self.session = requests.Session()
        self.session.auth = auth
        self.session.verify = verify
        self.session.cert = cert
        self.session.headers.update({
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        })
        self.timeout = timeout
        self.encoder = encoder
        self.decoder = decoder
        self.debug = debug

    def route(self, *args):
        """
        Return endpoint object
        """
        return APIEndpoint(
            endpoint=self.base_url + '/' + ('/'.join(map(str, args))),
            session=self.session, timeout=self.timeout,
            encoder=self.encoder, decoder=self.decoder, debug=self.debug
        )


class APIEndpoint(object):
    """
    API endpoint
    """
    def __init__(self, endpoint, session, timeout=None,
                 encoder=json.dumps, decoder=json.loads, debug=False):
        self.endpoint = endpoint
        self.session = session
        self.timeout = timeout
        self.encoder = json.dumps
        self.decoder = json.loads
        self.debug = debug
        self.last_response = None
    # ---

    def route(self, *args):
        """
        Return endpoint object
        """
        return APIEndpoint(
            endpoint=self.endpoint + '/' + ('/'.join(map(str, args))),
            session=self.session, timeout=self.timeout,
            encoder=self.encoder, decoder=self.decoder, debug=self.debug
        )
    # ---

    def do(self, method, params=None, **kwargs):
        """
        Do the HTTP request
        """
        method = method.upper()

        if self.debug:
            return dict(endpoint=self.endpoint, method=method, kwargs=kwargs, session=self.session)

        if method == 'GET':
            response = self.session.get(self.endpoint,
                                        params=params,
                                        timeout=self.timeout,
                                        **kwargs)
        else:
            response = self.session.request(method, self.endpoint,
                    data=self.encoder(params), params=params, timeout=self.timeout, **kwargs)
            
        content = response.content.decode('UTF-8')
        if response.status_code not in range(200,300):
            raise HTTPError(response)

        try:
            return self.decoder(content)
        except Exception:
            return content
    # ---

    def get(self, **kwargs): return self.do('GET', **kwargs)
    def post(self, **kwargs): return self.do('POST', **kwargs)
    def put(self, **kwargs): return self.do('PUT', **kwargs)
    def patch(self, **kwargs): return self.do('PATCH', **kwargs)
    def delete(self, **kwargs): return self.do('DELETE', **kwargs)



from .ensemblclient import EnsemblClient, _api_table
from .core import HTTPError, InvalidResponseError, _endpoint_docstring
from .assemblymapper import AssemblyMapper
from .utils import region_str

_default_client = EnsemblClient()

for endpoint in _api_table:
    docstring = _endpoint_docstring(endpoint)
    exec(f"def {endpoint['name']}(*args, **kwargs):                 \n"
         f"    '''                                                  \n"
         f"    {docstring}                                          \n"
         f"    '''                                                  \n"
         f"    return _default_client.{endpoint['name']}(*args, **kwargs)"
    )


from intervaltree import IntervalTree

from .ensemblclient import EnsemblClient
from .utils import region_str


class AssemblyMapper:
    """Structure that optimizes the mapping between 
    diferent genome assemblies.
    
    Example::
    
        >>> mapper = AssemblyMapper(from_assembly='GRCh37'
        ...                         to_assembly='GRCh38')
        
        >>> mapper.map(chrom='1', pos=1000000)
        1064620
        
    """
    # No need to request several times the same info
    map_cache = dict()
    
    def __init__(self, from_assembly='GRCh37', 
                       to_assembly='GRCh38', 
                       species='human'):
        """
        Parameters
        ----------
        
        from_assembly: str, optional
                Version of the input assembly. Default 'GRCh37'.
                
        to_assembly: str, optional
                Version of the output assembly. Default 'GRCh38'.
                
        species: str, optional
                Species name/alias. Default 'human'.
        """
        # The interface to the Ensembl REST API
        self.client = EnsemblClient()
        
        # The assembly mapping is formed by mappings 
        # of genome slices. We keep at minimum the 
        # number of requests we make by fetching the 
        # whole mapping information at once and storing 
        # it in an interval tree data structure for 
        # efficient lookup.
        self.assembly_map = self._fetch_mapping(from_assembly, 
                                                to_assembly,
                                                species)
    # ---
    
    def _fetch_mapping(self, from_assembly, to_assembly, species):
        """Fetch the mapping, either make it anew or use a cached one."""
        map_key = from_assembly + to_assembly + species
        
        if map_key in self.map_cache:
            # Mapping information already fetched, return it
            return self.map_cache[map_key]
        else:
            # Need to assemble the mapping information
            mapping = self._assemble_mapping(from_assembly, 
                                             to_assembly,
                                             species)
            # Cache the assembled mapping
            self.__class__.map_cache[map_key] = mapping
            
            return mapping
    # ---
    
    def _assemble_mapping(self, from_assembly, to_assembly, species):
        """Assemble the interval tree with the information 
        for the given mapping.
        """
        client = self.client

        # Get the chromosome sizes
        genome_info = client.assembly_info(species=species)
        chromosomes = {item['name']: item['length'] 
                            for item in genome_info['top_level_region']
                            if item['coord_system'] == 'chromosome'} 

        # Initialize the mapping structure
        # (An interval tree for each chromosome)
        assembly_map = dict()

        for chrom, chrom_len in chromosomes.items():
            # Query the mapping information
            # for the entire chromosome
            region = region_str(chrom, start=1, end=chrom_len)
            map_ = client.assembly_map(species=species,
                                       region=region,
                                       asm_one=from_assembly, 
                                       asm_two=to_assembly)

            # Assemble into the mapping structure
            assembly_map[chrom] = self._interval_tree(map_['mappings'], 
                                                      chrom_len)

        return assembly_map
    # ---
            
    def _interval_tree(self, mappings, chrom_length):
        """Assemble an interval tree from the mapping information.
        
        An interval tree is a tree data structure that allows to 
        efficiently find all intervals that overlap with any given 
        interval or point, often used for windowing queries.
            (see https://en.wikipedia.org/wiki/Interval_tree)
        
        The mapping information is a collection where 
        each item has the shape::
            {'mapped': 
               {'assembly': 'GRCh38',
                'coord_system': 'chromosome',
                'end': 1039365,
                'seq_region_name': 'X',
                'start': 1039265,
                'strand': 1},
             'original': 
               {'assembly': 'GRCh37',
                'coord_system': 'chromosome',
                'end': 1000100,
                'seq_region_name': 'X',
                'start': 1000000,
                'strand': 1}}
        """
        interval_tree = IntervalTree()
        
        for item in mappings:
            # Assemble the interval tree.
            # Each item describes a mapping of 
            # regions btw both assemblies.
            from_ = item['original']
            to = item['mapped']

            # Need to modify to represent a half open
            # interval (as [a,b) instead of [a,b])
            from_region = from_['start'],from_['end']+1

            if to['strand'] == +1:
                to_region = to['start'],to['end']
            else:
                # Handle mappings to the reverse strand
                # (Translate them to the forward strand)
                # Visual aid to the transformation:
                #  1  2  3  4  5  6  7  8  9 10
                #  |  |  |  |  |  |  |  |  |  |
                # 10 9  8  7  6  5  4  3  2  1
                to_region = (chrom_length - to['end'] + 1, 
                             chrom_length - to['start'] + 1)
            
            interval_tree.addi(*from_region, 
                               data=to_region)
        return interval_tree
    # ---
    
    def map(self, chrom, pos):
        """Map the given position.
        
        The mapping is between the specified assemblies
        when creating the object. (default: map position 
        from assembly GRCh37 to assembly GRCh38)
        """
        # Query the interval it maps to
        chromosome_mapper = self.assembly_map[str(chrom)]
        interval = chromosome_mapper[pos]

        if interval:
            # Calculate position from the interval
            interval = interval.pop()
            mapped_pos = interval.data[0] + (pos - interval.begin)
        else:
            # No mapping found.
            mapped_pos = None
        
        return mapped_pos
    # ---
# AssemblyMapper
